import uuid
from datetime import datetime
from typing import List

from kafka import KafkaProducer
from kafka.errors import NoBrokersAvailable
from streaming_data_types.run_stop_6s4t import serialise_6s4t as serialise_stop

from file_writer_control.CommandChannel import CommandChannel
from file_writer_control.CommandHandler import CommandHandler
from file_writer_control.CommandStatus import CommandStatus
from file_writer_control.JobStatus import JobState, JobStatus
from file_writer_control.KafkaTopicUrl import KafkaTopicUrl
from file_writer_control.WorkerStatus import WorkerStatus
from file_writer_control.WriteJob import WriteJob


class WorkerFinderBase:
    def __init__(
        self,
        command_topic: str,
        command_channel: CommandChannel,
        message_producer: KafkaProducer,
    ):
        """
        Constructor.
        """
        self.command_channel = command_channel
        self.command_topic = command_topic
        self.message_producer = message_producer

    def send_command(self, message: bytes):
        """
        Send a message (command) to the file-writer "command"-topic.
        :param message: The command/message as binary data.
        """
        self.message_producer.send(self.command_topic, message)

    def try_start_job(self, job: WriteJob) -> CommandHandler:
        """
        Attempts to start a file-writing job. This function is not blocking. No guarantees are given that the job will
        be successfully started.
        .. note:: This class must be implemented by the classes inheriting from this one.
        :param job: The file-writing job to be started.
        :return: A CommandHandler instance for (more) easily checking the outcome of attempting to start a write job.
        """
        raise NotImplementedError("Not implemented in base class.")

    def try_send_stop_time(
        self, service_id: str, job_id: str, stop_time: datetime
    ) -> CommandHandler:
        """
        Sends a "set stop time" message to a file-writer running a job as identified by the parameters.
        This function is not blocking. No guarantees are given that this command will be followed.
        :param service_id: The service identifier of the file-writer to receive the command.
        :param job_id: The job identifier of the currently running file-writer job.
        :param stop_time: The new stop time.
        :return: A CommandHandler instance for (more) easily checking the outcome of setting a new stop time.
        """
        command_id = str(uuid.uuid1())
        message = serialise_stop(
            job_id=job_id,
            service_id=service_id,
            command_id=command_id,
            stop_time=stop_time,
        )
        self.command_channel.add_command_id(job_id=job_id, command_id=command_id)
        self.send_command(message)
        return CommandHandler(self.command_channel, command_id)

    def try_send_stop_now(self, service_id: str, job_id: str) -> CommandHandler:
        """
        Sends a "stop now" message to a file-writer running a job as identified by the parameters of this function.
        This function is not blocking. No guarantees are given that this command will be followed.
        :param service_id: The service identifier of the file-writer to receive the command.
        :param job_id: The job identifier of the currently running file-writer job.
        :return: A CommandHandler instance for (more) easily checking the outcome of the "stop now" command.
        """
        command_id = str(uuid.uuid1())
        message = serialise_stop(
            job_id=job_id, service_id=service_id, command_id=command_id, stop_time=0
        )
        self.command_channel.add_command_id(job_id=job_id, command_id=command_id)
        self.send_command(message)
        return CommandHandler(self.command_channel, command_id)

    def list_known_workers(self) -> List[WorkerStatus]:
        """
        :return: A list of the (known) status of the workers publishing status updates to the configured command topic.
        """
        return self.command_channel.list_workers()

    def list_known_jobs(self) -> List[JobStatus]:
        """
        :return: A list of the (known) jobs and their status as published on the configured command topic.
        """
        return self.command_channel.list_jobs()

    def list_known_commands(self) -> List[CommandStatus]:
        """
        :return: A list of the (known) commands and their outcomes as published on the configured command topic.
        """
        return self.command_channel.list_commands()

    def get_job_state(self, job_id: str) -> JobState:
        """
        Get the state of a specific job.
        :param job_id: The (unique) identifier of the job that we are trying to find the state of.
        :return: The state of the job if known, JobState.UNAVAILABLE if job is not known.
        """
        current_job = self.command_channel.get_job(job_id)
        if current_job is None:
            return JobState.UNAVAILABLE
        return current_job.state

    def get_job_status(self, job_id: str) -> JobStatus:
        """
        Get the full (known) status of a specific job.
        :param job_id: The (unique) identifier of the job that we are trying to find the status of.
        :return: The status of the job if known. None if it is not.
        """
        return self.command_channel.get_job(job_id)


class WorkerFinder(WorkerFinderBase):
    def __init__(self, command_topic_url: str):
        temp_cmd_ch = CommandChannel(command_topic_url)
        command_url = KafkaTopicUrl(command_topic_url)
        try:
            temp_producer = KafkaProducer(bootstrap_servers=[command_url.host_port])
        except NoBrokersAvailable as e:
            raise NoBrokersAvailable(
                f'Unable to find brokers (or connect to brokers) on address: "{command_url.host_port}"'
            ) from e
        super().__init__(command_url.topic, temp_cmd_ch, temp_producer)



import atexit
import threading
import time
from random import randrange
from typing import List

from file_writer_control.CommandHandler import CommandHandler
from file_writer_control.JobStatus import JobState
from file_writer_control.WorkerFinder import WorkerFinder
from file_writer_control.WorkerStatus import WorkerState, WorkerStatus
from file_writer_control.WriteJob import WriteJob

START_JOB_TIMEOUT = 30  # Seconds
SEND_JOB_TIMEOUT = 10  # Seconds


class WorkerCommandChannel(WorkerFinder):
    """
    A child of WorkerFinder intended for use with "direct command" style of starting a file-writing job.
    """

    def __init__(self, command_topic_url: str):
        super().__init__(command_topic_url)
        self.start_job_threads = []
        self._stop_event = threading.Event()

        def do_exit():
            self.stop_threads()

        atexit.register(do_exit)

    def stop_threads(self):
        self._stop_event.set()
        for thread in self.start_job_threads:
            thread.join()

    def try_start_job(self, job: WriteJob) -> CommandHandler:
        """
        Attempt to start a file-writing job by finding an available (idle) worker and sending it a write command.
        Will re-try to send a write command until a timeout has been reached. This member function is not blocking.
        :param job: The job to be started.
        :return: A CommandHandler for (more) easily checking the outcome of the start command.
        """
        thread_kwargs = {"do_job": job, "stop_event": self._stop_event}
        self.command_channel.add_job_id(job.job_id)
        self.command_channel.add_command_id(job.job_id, job.job_id)
        temp_thread = threading.Thread(
            target=self._start_job_thread_function, daemon=True, kwargs=thread_kwargs
        )
        temp_thread.start()
        self.start_job_threads.append(temp_thread)
        return CommandHandler(self.command_channel, job.job_id)

    def get_idle_workers(self) -> List[WorkerStatus]:
        """
        Get a list of workers that has the (known) state idle. May be an empty list. Idle status may be outdated.
        :return: List of workers.
        """
        list_of_workers = self.list_known_workers()
        list_of_idle_workers = []
        for worker in list_of_workers:
            if worker.state == WorkerState.IDLE:
                list_of_idle_workers.append(worker)
        return list_of_idle_workers

    def _start_job_thread_function(self, do_job: WriteJob, stop_event: threading.Event):
        """
        Attempt to find an idle worker (file-writer) and give it a file-writing job.
        :param do_job: The write job to start.
        :param stop_event: Used for stopping the thread before it has finished by itself.
        """
        job_started_time = start_time = time.time()

        waiting_to_send_job = True
        loop_poll_rate = 1.0  # Hz
        while start_time + START_JOB_TIMEOUT > time.time() and not stop_event.is_set():
            if waiting_to_send_job:
                list_of_idle_workers = self.get_idle_workers()
                if len(list_of_idle_workers) > 0:
                    used_worker = list_of_idle_workers[
                        randrange(len(list_of_idle_workers))
                    ]
                    do_job.service_id = used_worker.service_id
                    self.message_producer.send(
                        self.command_topic, do_job.get_start_message()
                    )
                    waiting_to_send_job = False
                    job_started_time = time.time()
            else:
                list_of_jobs = self.command_channel.list_jobs()
                for job in list_of_jobs:
                    if job.job_id == do_job.job_id:
                        if job.state == JobState.WRITING or job.state == JobState.DONE:
                            return
                        elif job_started_time + SEND_JOB_TIMEOUT < time.time():
                            waiting_to_send_job = True
                            do_job.generate_new_job_id()
                        break
            time.sleep(1.0 / loop_poll_rate)



from ._version import version
from .JobHandler import JobHandler
from .WorkerCommandChannel import WorkerCommandChannel
from .WorkerJobPool import WorkerJobPool
from .WriteJob import WriteJob
from .CommandStatus import CommandState
from .JobStatus import JobState

__all__ = [
    "JobHandler",
    "WorkerCommandChannel",
    "WorkerJobPool",
    "WriteJob",
    "CommandState",
    "JobState",
    "version",
]



from datetime import datetime, timedelta
from enum import Enum, auto

JOB_STATUS_TIMEOUT = timedelta(seconds=5)


class JobState(Enum):
    """
    The state of a job.
    """

    NO_JOB = auto()
    WAITING = auto()
    WRITING = auto()
    TIMEOUT = auto()
    ERROR = auto()
    DONE = auto()
    UNAVAILABLE = auto()


class JobStatus:
    """
    Contains general information about the (execution) of a job.
    """

    def __init__(self, job_id: str):
        self._job_id = job_id
        self._service_id = ""
        self._file_name = ""
        self._last_update = datetime.now()
        self._state = JobState.WAITING
        self._message = ""

    def update_status(self, new_status: "JobStatus") -> None:
        """
        Updates the status/state of a this instance of the JobStatus class, using another instance.
        .. note:: The job identifier of this instance and the other must be identical.
        :param new_status: The other instance of the JobStatus class.
        """
        if new_status.job_id != self.job_id:
            raise RuntimeError(
                f"Job id of status update is not correct ({self.job_id} vs {new_status.job_id})"
            )
        self._state = new_status.state
        if new_status.message:
            self._message = new_status.message
        self._service_id = new_status.service_id
        self._file_name = new_status.file_name
        self._last_update = new_status.last_update

    def check_if_outdated(self, current_time: datetime):
        """
        Given the current time, state and the time of the last update: Have we lost the connection?
        :param current_time: The current time
        """
        if (
            self.state != JobState.DONE
            and self.state != JobState.ERROR
            and self.state != JobState.TIMEOUT
            and current_time - self.last_update > JOB_STATUS_TIMEOUT
        ):
            self._state = JobState.TIMEOUT
            self._last_update = current_time

    @property
    def job_id(self) -> str:
        """
        The (unique) job identifier.
        """
        return self._job_id

    @property
    def service_id(self) -> str:
        """
        The (unique) service identifier of the instance of the file-writer that executes the current job.
        """
        return self._service_id

    @service_id.setter
    def service_id(self, new_service_id: str) -> None:
        if not self._service_id:
            self._service_id = new_service_id
            self._last_update = datetime.now()
        elif self._service_id == new_service_id:
            return
        else:
            raise RuntimeError(
                f'Can not set service_id of job with id "{self._job_id}" to "{new_service_id}" as it has already been set to "{self._service_id}".'
            )

    @property
    def last_update(self) -> datetime:
        """
        The local time stamp of the last update of the status of the job.
        """
        return self._last_update

    @property
    def state(self) -> JobState:
        """
        The current state of the job.
        """
        return self._state

    @property
    def file_name(self) -> str:
        """
        The file name of the job. None if the file name is not known.
        """
        if self._file_name == "":
            return None
        return self._file_name

    @file_name.setter
    def file_name(self, new_file_name: str) -> None:
        self._file_name = new_file_name
        self._last_update = datetime.now()

    @state.setter
    def state(self, new_state: JobState) -> None:
        self._state = new_state
        self._last_update = datetime.now()

    @property
    def message(self) -> str:
        """
        Status/state message of the job as received from the file-writer.
        """
        return self._message

    @message.setter
    def message(self, new_message: str) -> None:
        if new_message:
            self._message = new_message
            self._last_update = datetime.now()
